#!/usr/bin/python
import subprocess
import os
import re
import time
from urlparse import urlparse

from charmhelpers import fetch
from charmhelpers.core import hookenv, host
from charmhelpers.contrib import ssl

from apache_helpers import ApacheVhostRelation

conf_dir = '/etc/apache2/'
cert_dir = os.path.join(conf_dir, 'ssl/')
vhost_filename = 'subordinate-vhost.conf'
vhost_path = os.path.join(conf_dir, 'sites-available', vhost_filename)


def install_apache(service_name):
    """ This callback is scheduled to run on just about every event.  If the
    mpm config value has changed, it will install the appropriate apache
    packages. """

    config = hookenv.config()
    mpm = config['mpm_type']
    release = host.lsb_release()['DISTRIB_RELEASE']

    if config.changed('mpm_type'):
        packages = ['apache2', 'libapache2-mod-apparmor']

        fetch.apt_install(packages)

        prev_mpm = subprocess.check_output(('/usr/sbin/a2query', '-M',)).strip()

        # if the default mpm used by the apache2 package is the same as the
        # default mpm_type in the config.yaml, config.changed('mpm_type') will
        # be true on the first run of this function and prev_mpm will be equal
        # to mpm, which will lead to the purge/disable of the only mpm module
        # below. We want to avoid, so check it.
        if prev_mpm != mpm:

            # if before xenial, this is all done with packages
            if release < "16.04":
                fetch.apt_install(("apache2-mpm-{}".format(mpm),))
                if prev_mpm:
                    fetch.apt_purge(("apache2-mpm-{}".format(prev_mpm),))
            # starting from xenial, all the mpm modules are in the apache2 package
            # and need to be handled via a2{en,dis}mod.
            else:
                subprocess.check_call(('/usr/sbin/a2dismod', 'mpm_{}'.format(prev_mpm)))
                subprocess.check_call(('/usr/sbin/a2enmod', 'mpm_{}'.format(mpm)))

            host.service_restart("apache2")


def write_certs(service_name):
    """ There are three base64-encoded config options that may contain PEM
    blobs for the apache SSl configuration.  If we find them suitable, we write
    them to files named after their CN. Then each relation either makes use of
    them or generates a snakeoil cert. """

    from base64 import b64decode

    config = hookenv.config()
    subject = {'CN': None, 'Alt': []}
    host.mkdir(cert_dir, perms=0o500)
    if config['ssl_certificate'] and config['ssl_certificate'] != 'AUTOCERT':
        cert = b64decode(config['ssl_certificate'])
        x509_process = subprocess.Popen(
            ('/usr/bin/openssl', 'x509', '-noout', '-text'),
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,)
        output, _ = x509_process.communicate(cert)
        subject['CN'] = re.search(" *Subject: .*CN=(.*)", output).group(1)
        SANs = re.findall("DNS:([^ ,]+)", output)
        subject['Alt'] = [x.strip() for x in SANs if x != subject['CN']]
        hookenv.log('Received cert for CN: {CN}, SubjectAltNames: {Alt}'.format(**subject))

        host.write_file(
            os.path.join(cert_dir, '{CN}.crt'.format(**subject)),
            cert,
        )
        host.write_file(
            os.path.join(cert_dir, '{CN}.key'.format(**subject)),
            b64decode(config['ssl_key']),
            perms=0o400,
        )
        if config["ssl_chain"]:
            host.write_file(
                os.path.join(cert_dir, '{CN}-chain.pem'.format(**subject)),
                b64decode(config['ssl_chain']),
            )
        for SAN in subject['Alt']:
            if not os.path.exists(os.path.join(cert_dir, '{}.crt'.format(SAN))):
                os.symlink(os.path.join(cert_dir, '{CN}.crt'.format(**subject)),
                           os.path.join(cert_dir, '{}.crt'.format(SAN)))
            if not os.path.exists(os.path.join(cert_dir, '{}.key'.format(SAN))):
                os.symlink(os.path.join(cert_dir, '{CN}.key'.format(**subject)),
                           os.path.join(cert_dir, '{}.key'.format(SAN)))
            if (not os.path.exists(os.path.join(cert_dir, '{}-chain.pem'.format(SAN)))
                and config["ssl_chain"]):
                os.symlink(os.path.join(cert_dir, '{CN}-chain.pem'.format(**subject)),
                           os.path.join(cert_dir, '{}-chain.pem'.format(SAN)))

    vhosts = ApacheVhostRelation()['webservice']
    for unit in vhosts:
        for service in unit['services']:
            if service['url'].scheme != 'https':
                continue
            subprocess.check_call(("/usr/sbin/a2enmod", "ssl"))
            if config['ssl_certificate'] == 'AUTOCERT':
                continue
            if service['url'].hostname == subject['CN']:
                continue  # We already have the cert written out
            if service['url'].hostname in subject['Alt']:
                continue
            hookenv.log('Generating self-signed cert for {}'.format(
                service['url'].hostname))
            keyfile = os.path.join(cert_dir,
                                   '{CN}.key'.format(CN=service['url'].hostname))
            certfile = os.path.join(cert_dir,
                                    '{CN}.crt'.format(CN=service['url'].hostname))
            ssl.generate_selfsigned(keyfile, certfile,
                                    cn=service['url'].hostname)
            if config["ssl_chain"]:
                chainfile = os.path.join(cert_dir,
                                         '{CN}-chain.pem'.format(CN=service['url'].hostname))
                chainkeyfile = os.path.join(cert_dir,
                                            '{CN}-chain.key'.format(CN=service['url'].hostname))
                ssl.generate_selfsigned(chainkeyfile, chainfile,
                                        cn='Self Signed Chain')


def install_extra_packages(service_name):
    """ Included for historic reasons: the old charm had the superior charm
    pass packages to install down to the subordinate.  It's preferred that the
    superior charm install what it needs.  We do use this relation to see if
    PHP is needed, and to install the apache module and associated accelerator.
    """

    config = hookenv.config()
    vhosts = ApacheVhostRelation()['webservice']
    extra_packages = set()
    release = host.lsb_release()['DISTRIB_RELEASE']

    for unit in vhosts:
        for service in unit['services']:
            if 'extra_packages' in service:
                extra_packages.update(service['extra_packages'])
            if 'php' in service['type']:
                hookenv.log('PHP needed: installing accelerator and switching to'
                            ' prefork apache')
                if release < "16.04":
                    extra_packages.update(("libapache2-mod-php5", "php-apc"))
                else:
                    extra_packages.update(("libapache2-mod-php", "php-apcu"))
                config['mpm_type'] = 'prefork'
                config.save()
        fetch.apt_install(extra_packages)
        host.service_restart("apache2")


def enable_vhost(service_name):
    if host.is_container():
        subprocess.check_call(('/usr/sbin/a2dismod', 'apparmor'))
    else:
        if os.path.exists("/etc/apparmor.d/disable/usr.sbin.apache2"):
            os.unlink("/etc/apparmor.d/disable/usr.sbin.apache2")
            host.service_reload('apparmor')
        subprocess.check_call(('/usr/sbin/a2enmod', 'apparmor'))
    subprocess.check_call(('/usr/sbin/a2ensite', vhost_filename))
    host.service_restart("apache2")


def disable_vhost(service_name):
    subprocess.check_call(('/usr/sbin/a2dissite', vhost_filename))
    host.service_restart("apache2")


def vhost_ready(service_name):
    """Notify other side of the relation that vhosts are configured."""
    for rel_id in hookenv.relation_ids('webservice'):
        timestamp = int(time.time())
        hookenv.relation_set(relation_id=rel_id, vhost_ready=timestamp)


def _get_proxy_modules(services):
    """Return set of apache modules required to handle ProxyPass requests.
    If no proxy is needed returns empty set
    """
    from itertools import chain
    proxies = chain.from_iterable(service['proxy'] for service in services if 'proxy' in service)
    modules = []
    for proxy in proxies:
        if urlparse(proxy['proxy_target']).scheme == 'https':
            modules.append('proxy_http')
        else:
            modules.append('proxy_' + urlparse(proxy['proxy_target']).scheme)
    if modules:
        modules.append('proxy')
        hookenv.log('Proxy in use, adding required modules: {}'.format(set(modules)))
    return set(modules)


def enable_disable_options(service_name):
    """ Web server options for Apache are distinguishable as modules in
    /etc/apache2/mods-available/ and config file stubs in
    /etc/apache2/conf-available/.  We accept an iterable of strings with the
    following rules:

        mod_foo     indicates that the option is a module
        foo.conf    indicates that the option is a config file

    In addition a '-' prefix indicates that we explicitly wish to disable the
    option, and an optional '+' prefix is stripped off and ignored when
    enabling the option.

    Thus the following tuple would enable the foo and bar modules, disable the
    baz module, enable the foo and bar configs, and disable the baz config:

        ('mod_foo', '+mod_bar', '-mod_baz',
         'foo.conf', '+bar.conf', '-baz.conf')

    Any previously enabled option not specified by any service will be left
    alone.
    """
    from itertools import chain

    vhosts = ApacheVhostRelation()['webservice']
    options = set()

    services = []
    for unit in vhosts:
        services.extend(unit['services'])
    options.update(chain.from_iterable(service['webserver_options'] for service
                   in services if 'webserver_options' in service))

    mods_enabled = {x.partition('mod_')[2] for x in options if
                    x.startswith(('+mod_', 'mod_'))}
    mods_enabled.update(_get_proxy_modules(services))
    mods_disabled = {x.partition('mod_')[2] for x in options if
                     x.startswith('-mod_')}
    confs_enabled = {x.rpartition('.conf')[0].lstrip('+') for x in options if
                     x.endswith('.conf') and not x.startswith('-')}
    confs_disabled = {x.rpartition('.conf')[0].lstrip('-') for x in options if
                      x.endswith('.conf') and x.startswith('-')}

    if mods_enabled.intersection(mods_disabled):
        raise RuntimeError('Modules conflict between services: {}'.format(
            mods_enabled.intersection(mods_disabled)))

    if confs_enabled.intersection(confs_disabled):
        raise RuntimeError('Config file conflict between services: {}'.format(
            confs_enabled.intersection(confs_disabled)))

    for module in mods_enabled:
        subprocess.check_call(('/usr/sbin/a2enmod', module))

    for module in mods_disabled:
        subprocess.check_call(('/usr/sbin/a2dismod', module))

    for conffile in confs_enabled:
        subprocess.check_call(('/usr/sbin/a2enconf', conffile))

    for conffile in confs_disabled:
        subprocess.check_call(('/usr/sbin/a2disconf', conffile))

    host.service_restart("apache2")
